// Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
//
// This software may be used and distributed according to the terms of the
// GNU General Public License version 2 or any later version.
// SPDX-License-Identifier: GPL-2.0-or-later
use clap::Parser;
use std::error::Error;
use std::net::SocketAddr;
use std::path::PathBuf;
use std::sync::Arc;

use tokio::fs::remove_file;
use tokio::net::UnixListener;
use tokio_stream::wrappers::UnixListenerStream;
use tonic::transport::Server;

use rhgitaly::config::Config;
use rhgitaly::service::blob::blob_server;
use rhgitaly::service::commit::commit_server;
use rhgitaly::service::mercurial_repository::mercurial_repository_server;
use rhgitaly::service::r#ref::ref_server;
use rhgitaly::service::repository::repository_server;
use rhgitaly::service::server::{server_server, HGITALY_VERSION};
use rhgitaly::streaming::WRITE_BUFFER_SIZE;
use tracing::{info, Level};
use tracing_subscriber::FmtSubscriber;

fn setup_tracing() {
    // a builder for `FmtSubscriber`.
    let subscriber = FmtSubscriber::builder()
        .with_max_level(Level::INFO)
        .finish();

    tracing::subscriber::set_global_default(subscriber).expect("setting default subscriber failed");
}

#[derive(Debug)]
#[allow(clippy::upper_case_acronyms)]
enum BindAddress {
    Unix(PathBuf),
    TCP(SocketAddr),
}

/// Parse the URL to listen to
///
/// There are two cases: `unix:/some/path/sock` and `tcp://hostname:port`, where
/// hostname can only be an IP address (v4 or v6).
///
/// This is a toy implementation anyway, we'll add a proper dependency to the `url` crate later on
fn parse_listen_url(url: &str) -> Result<BindAddress, Box<dyn Error>> {
    if let Some(path) = url.strip_prefix("unix:") {
        return Ok(BindAddress::Unix(path.into()));
    }
    if let Some(addr) = url.strip_prefix("tcp://") {
        return Ok(BindAddress::TCP(addr.parse()?));
    }
    Err("Unsupported URL".into())
}

#[derive(Parser, Debug)]
#[command(name = "RHGitaly")]
#[command(
    about = "RHGitaly is a partial implementation of the Gitaly and HGitaly gRPC protocols \
for Mercurial repositories"
)]
#[command(
    long_about = "RHGitaly is a performance-oriented partial implementation of the Gitaly \
and HGitaly gRPC protocols for Mercurial repositories.

It is asynchronous with a pool of worker threads, leveraging the Tonic (Tokio) gRPC framework, \
and the Mercurial primitives implemented in Rust (hg-core crate).

Configuration is for now entirely done by environment variables (see
https://foss.heptapod.net/heptapod/hgitaly/-/issues/181 to follow progress on this)
"
)]
#[command(version = HGITALY_VERSION)]
struct Args {}

use tokio::signal::unix::{signal, SignalKind};
use tokio_util::sync::CancellationToken;

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    let _args = Args::parse();

    setup_tracing();

    info!("RHGitaly starting, version {}", HGITALY_VERSION);
    info!("WRITE_BUFFER_SIZE={}", *WRITE_BUFFER_SIZE);

    let shutdown_token = CancellationToken::new();

    let config = Arc::new(Config::from_env());
    let server = Server::builder()
        .add_service(server_server())
        .add_service(blob_server(&config))
        .add_service(commit_server(&config))
        .add_service(mercurial_repository_server(&config, &shutdown_token))
        .add_service(ref_server(&config))
        .add_service(repository_server(&config));

    let bind_addr = parse_listen_url(&config.listen_url)?;
    info!("RHGitaly binding to {:?}", &bind_addr);

    let mut sigterm = signal(SignalKind::terminate())?;
    // sigterm.recv() returns `None` if "no more events can be received by this stream"
    // In the case of SIGTERM, it probably cannot happen, and by default we will consider
    // it to be a termination condition.
    let wait_sigterm = async move {
        sigterm.recv().await;
        shutdown_token.cancel();
    };

    match bind_addr {
        BindAddress::TCP(addr) => server.serve_with_shutdown(addr, wait_sigterm).await?,
        BindAddress::Unix(ref path) => {
            let uds = UnixListener::bind(path)?;
            let uds_stream = UnixListenerStream::new(uds);
            server
                .serve_with_incoming_shutdown(uds_stream, wait_sigterm)
                .await?
        }
    };

    // Server now has shut down
    if let BindAddress::Unix(path) = bind_addr {
        info!("Removing Unix Domain socket file at '{}'", &path.display());
        remove_file(&path).await?
    }

    Ok(())
}
